mod constants;
pub mod token_generator;

use isahc::auth::{Authentication, Credentials};
use isahc::http::HeaderMap;
use isahc::prelude::*;
use isahc::{Body, Request, Response};
use std::str::FromStr;
use std::sync::Arc;

#[allow(unused_imports)]
use log::*;

struct Proxy {
    address: String,
    credentials: Option<Credentials>,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum ProxyError {
    InvalidProxyFormat,
}

impl FromStr for Proxy {
    type Err = ProxyError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut parts = s.split(':');
        let host = parts.next().ok_or(ProxyError::InvalidProxyFormat)?;
        let port = parts.next().ok_or(ProxyError::InvalidProxyFormat)?;
        let auth = if let Some(user) = parts.next() {
            let pass = parts.next().ok_or(ProxyError::InvalidProxyFormat)?;
            Some(Credentials::new(user, pass))
        } else {
            None
        };
        Ok(Proxy {
            address: format!("{}:{}", host, port),
            credentials: auth,
        })
    }
}

#[derive(Clone)]
pub struct Client {
    http_client: Arc<isahc::HttpClient>,
    proxy: Option<Arc<Proxy>>,
    twitter_auth: Option<TwitterAuth>,
}

#[derive(Clone)]
pub struct TwitterAuth {
    bearer: String,
    guest_id: Option<String>,
}

impl Default for Client {
    fn default() -> Self {
        Self::new()
    }
}

impl Client {
    pub fn new() -> Self {
        Client {
            http_client: Arc::new(isahc::HttpClient::new().expect("failed to create http client")),
            proxy: None,
            twitter_auth: None,
        }
    }

    pub fn new_with_proxy(proxy: &str) -> Self {
        let proxy = proxy.parse().unwrap();
        Client {
            http_client: Arc::new(isahc::HttpClient::new().expect("failed to create http client")),
            proxy: Some(Arc::new(proxy)),
            twitter_auth: None,
        }
    }

    fn get(&self, url: &str, headers: &HeaderMap) -> Result<Response<Body>, isahc::Error> {
        let mut request = Request::get(url);
        if let Some(proxy) = &self.proxy {
            request = request.proxy(Some(proxy.address.parse().expect("error parsing proxy!")));
            if let Some(creds) = &proxy.credentials {
                request = request.proxy_authentication(Authentication::basic());
                request = request.proxy_credentials(creds.clone());
            }
        }
        let mut request = request.body(())?;
        request
            .headers_mut()
            .extend(headers.iter().map(|(k, v)| (k.clone(), v.clone())));
        self.http_client.send(request)
    }
}

#[cfg(test)]
mod tests {
    use crate::token_generator::get_bearer_and_guest_stuff;

    use super::*;

    #[test]
    fn world_normal() {
        let result = 2 + 2;
        assert_eq!(result, 4);
    }

    #[test]
    fn get_bearer_token() {
        env_logger::init();
        let proxy =
            std::env::var("TESTING_PROXY").unwrap_or_else(|_| panic!("use a proxy for testing!!!"));
        let client = Client::new_with_proxy(proxy.as_str());
        let (bearer, guest_id, guest_token) = get_bearer_and_guest_stuff(&client).unwrap();
        debug!("bearer: {}", bearer);
        debug!("guest_id: {}", guest_id);
        debug!("guest_token: {}", guest_token);
    }
}
